package com.emamaker.voxeltest.intervaltrees.renderer;

import java.util.ArrayDeque;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Queue;

import com.emamaker.voxeltest.intervaltrees.data.IntervalTree;
import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.world.Chunk;
import com.emamaker.voxeltest.intervaltrees.world.WorldManager;
import com.emamaker.voxeltest.intervaltrees.world.blocks.Blocks;
import com.jme3.material.Material;
import com.jme3.math.ColorRGBA;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.shape.Box;
import com.jme3.scene.shape.Sphere;

public class ChunkRenderer {

    // Just render a full cube for every block
    public void stupidMeshing(WorldManager mgr, Chunk chunk) {
        // Breadth-first visit each node of the tree
        chunk.blocks.print();

        Queue<IntervalTree<Blocks>.TreeNode> queue = new ArrayDeque();
        queue.add(chunk.blocks.getRoot());

        IntervalTree<Blocks>.TreeNode t;
        Box b;
        Geometry geom;
        Material mat;
        int[] coords1, coords2;

        while (!queue.isEmpty()) {
            t = queue.poll();
            if (t.getLeft() != null) queue.add(t.getLeft());
            if (t.getRight() != null) queue.add(t.getRight());

            if (!t.getValue().equals(Blocks.AIR)) {

                int start = t.getInterval().getLow();
                for (int i = t.getInterval().getLow(); i <= t.getInterval().getHigh(); i++) {

                    if ((i % Config.CHUNK_SIZE == 0 && i != t.getInterval().getLow()) || i == t.getInterval().getHigh()) {
                        if (i == t.getInterval().getHigh()) i++;
                        b = new Box((i - start) * 0.5f, 0.5f, 0.5f);

                        geom = new Geometry("Box", b);
                        mat = new Material(mgr.game.getAssetManager(), "Common/MatDefs/Misc/Unshaded.j3md");
                        mat.getAdditionalRenderState().setWireframe(true);
                        geom.setMaterial(mat);

                        coords1 = Chunk.coord1DTo3D(start);
                        coords2 = Chunk.coord1DTo3D(i);

                        geom.getLocalTransform().setTranslation(coords1[0] + b.xExtent, coords1[1] + b.yExtent, coords1[2] + b.zExtent);
                        System.out.println(Arrays.toString(coords1) + "-" + Arrays.toString(coords2) + "\t\tBox with extext " + b.xExtent + " (both directions) with center"
                                + geom.getLocalTransform().getTranslation());

                        chunk.chunkNode.attachChild(geom);

                        if (t.getValue() == Blocks.DIRT) {
                            mat.setColor("Color", ColorRGBA.Brown);
                        } else if (t.getValue() == Blocks.STONE) {
                            mat.setColor("Color", ColorRGBA.Gray);
                        } else if (t.getValue() == Blocks.GRASS) {
                            mat.setColor("Color", ColorRGBA.Green);
                        }

                        start = i;
                    }
                }

                // System.out.println("Placing a box long " + b.xExtent + " at "
                // + Arrays.toString(coords));

            }
            if (!mgr.game.getRootNode().hasChild(chunk.chunkNode)) mgr.game.getRootNode().attachChild(chunk.chunkNode);
        }
    }

    public void stupidArrayMeshing(WorldManager mgr, Chunk chunk) {
        Blocks[] blocks = chunk.treeTo1DArray();

        int idx;
        for (int i = 0; i < Config.CHUNK_SIZE; i++) {
            for (int j = 0; j < Config.CHUNK_SIZE; j++) {

                for (int k = 0; k < Config.CHUNK_SIZE; k++) {
                    idx = Chunk.coord3DTo1D(i, j, k);
                    if (blocks[idx] != Blocks.AIR) {
                        Box b = new Box(0.5f, 0.5f, 0.5f);

                        Geometry geom = new Geometry("Box", b);
                        Material mat = new Material(mgr.game.getAssetManager(), "Common/MatDefs/Misc/Unshaded.j3md");
                        // mat.getAdditionalRenderState().setWireframe(true);
                        geom.setMaterial(mat);
                        geom.getLocalTransform().setTranslation(i - 0.5f, j - 0.5f, k - 0.5f);

                        chunk.chunkNode.attachChild(geom);

                        if (blocks[idx] == Blocks.DIRT) {
                            mat.setColor("Color", ColorRGBA.Brown);
                        } else if (blocks[idx] == Blocks.STONE) {
                            mat.setColor("Color", ColorRGBA.Gray);
                        } else if (blocks[idx] == Blocks.GRASS) {
                            mat.setColor("Color", ColorRGBA.Green);
                        } else {
                            mat.setColor("Color", ColorRGBA.Red);
                        }
                    }
                }
            }
        }

    }

    public void greedyMeshing(WorldManager mgr, Chunk chunk) {
        HashMap<Vector3f, Sphere> points = new HashMap<>();

        // convert tree to array since it is easier to work with it
        Blocks[] blocks = chunk.treeTo1DArray();
        System.out.println(Arrays.toString(blocks));

        Vector3f pos, belowPos;
        int idx;

        for (int j = 0; j < 1; j++) {
            Blocks prevBlock = null;

            // For now I'm placing little spheres to show vertices instead of
            // actually constructing a mesh
            // Currently only meshing on XZ plane

            for (int k = 0; k < Config.CHUNK_SIZE; k++) {

                for (int i = 0; i < Config.CHUNK_SIZE; i++) {
                    idx = Chunk.coord3DTo1D(i, j, k);
                    pos = new Vector3f(i, j, k);
                    belowPos = new Vector3f(i, j, k - 1);

                    // Got a new type of block, place a vertex
                    if (blocks[idx] != prevBlock || prevBlock == null) {
                        // Check if we are at the start, if so just place
                        // vertices
                        if (k == 0) {
                            points.put(pos, new Sphere(25, 25, 0.25f));
                        }

                        // check if we are at the second step (right after
                        // start)
                        // if so, a new vertex needs to be placed whenever a
                        // quad can be completed
                        // if a quad cannot be completed but the block has
                        // changed, also place a vertex 1 lower to complete a
                        // quad
                        else if (k == 1) {
                            points.put(pos, new Sphere(25, 25, 0.25f));
                            if (blocks[idx] != blocks[Chunk.coord3DTo1D(i, j, k - 1)]) points.put(belowPos, new Sphere(25, 25, 0.25f));
                        }
                        // in any other situation
                        else {
                            if (blocks[idx] == blocks[Chunk.coord3DTo1D(i, j, k - 1)]) {
                                points.put(pos, new Sphere(25, 25, 0.25f));
                                points.remove(belowPos);
                                System.out.println("should be moving from " + belowPos + " to " + pos);
                            } else if (blocks[idx] != blocks[Chunk.coord3DTo1D(i, j, k - 1)]) points.put(belowPos, new Sphere(25, 25, 0.25f));
                        }
                    }
                    prevBlock = blocks[idx];
                }

            }
            for (Vector3f v : points.keySet()) {
                System.out.println(v);
                Geometry geom = new Geometry("Sphere", points.get(v));
                geom.getLocalTransform().setTranslation(v);

                Material mat = new Material(mgr.game.getAssetManager(), "Common/MatDefs/Misc/Unshaded.j3md");
                mat.getAdditionalRenderState().setWireframe(true);
                geom.setMaterial(mat);

                chunk.chunkNode.attachChild(geom);
            }

            points.clear();
        }

    }
}
