package com.emamaker.voxeltest.intervaltrees.world;

import java.util.ArrayDeque;
import java.util.Queue;

import com.emamaker.voxeltest.intervaltrees.data.Interval;
import com.emamaker.voxeltest.intervaltrees.data.IntervalTree;
import com.emamaker.voxeltest.intervaltrees.utils.Config;
import com.emamaker.voxeltest.intervaltrees.utils.Utils;
import com.emamaker.voxeltest.intervaltrees.world.blocks.Blocks;
import com.jme3.material.Material;
import com.jme3.math.Vector3f;
import com.jme3.scene.Geometry;
import com.jme3.scene.Node;

public class Chunk {

    Vector3f pos;
    public IntervalTree<Blocks> blocks = new IntervalTree<Blocks>();
    public Node chunkNode = new Node();

    public Chunk() {
        this(Vector3f.ZERO);
    }

    public Chunk(Vector3f pos_) {
        // I still have to decided if this is necessary. With an empty tree this
        // takes O(1)
        blocks.insertValue(Blocks.AIR, new Interval(0, Config.CHUNK_3DCOORD_MAX_INDEX));
    }

    public static int coord3DTo1D(int x, int y, int z) {
        return Utils.coord3DTo1D(x, y, z, Config.CHUNK_SIZE, Config.CHUNK_SIZE, Config.CHUNK_SIZE);
    }

    public static int[] coord1DTo3D(int idx) {
        return Utils.coord1DTo3D(idx, Config.CHUNK_SIZE, Config.CHUNK_SIZE, Config.CHUNK_SIZE);
    }

    public Blocks getBlock(int x, int y, int z) {
        return blocks.getSingleValue(Chunk.coord3DTo1D(x, y, z));
    }

    public void generatePlane() {
        // blocks.print();
        // blocks.insertValue(Blocks.STONE, new Interval(0,
        // Config.CHUNK_3DCOORD_MAX_INDEX));

        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.STONE, new Interval(coord3DTo1D(0, 0, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 0, i)));
        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.DIRT, new Interval(coord3DTo1D(0, 1, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 1, i)));
        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.GRASS, new Interval(coord3DTo1D(0, 2, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 2, i)));

        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.STONE, new Interval(coord3DTo1D(0, 3, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 3, i)));
        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.DIRT, new Interval(coord3DTo1D(0, 4, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 4, i)));
        for (int i = 0; i < Config.CHUNK_SIZE; i++) blocks.insertValue(Blocks.GRASS, new Interval(coord3DTo1D(0, 5, i), coord3DTo1D(Config.CHUNK_SIZE - 1, 5, i)));
    }

    public void generateStair() {
        for (int i = 0; i < Config.CHUNK_SIZE; i++) {
            for (int j = 0; j < Config.CHUNK_SIZE; j++) {
                blocks.insertValue(j < Config.CHUNK_SIZE - i - 1 ? Blocks.STONE : Blocks.GRASS, new Interval(coord3DTo1D(0, i, j), coord3DTo1D(Config.CHUNK_SIZE - 1, i, j)));
            }
        }

        blocks.insertValue(Blocks.DIRT, new Interval(Chunk.coord3DTo1D(0, 0, 0), Chunk.coord3DTo1D(2, 0, 0)));
        blocks.insertValue(Blocks.DIRT, new Interval(Chunk.coord3DTo1D(1, 0, 2), Chunk.coord3DTo1D(2, 0, 2)));
    }

    public Blocks[] treeTo1DArray() {
        // System.out.println(Config.CHUNK_3DCOORD_MAX_INDEX);
        Blocks[] result = new Blocks[Config.CHUNK_3DCOORD_MAX_INDEX + 1];

        Queue<IntervalTree<Blocks>.TreeNode> queue = new ArrayDeque();
        queue.add(blocks.getRoot());

        IntervalTree<Blocks>.TreeNode t;

        while (!queue.isEmpty()) {
            t = queue.poll();
            if (t.getLeft() != null) queue.add(t.getLeft());
            if (t.getRight() != null) queue.add(t.getRight());

            for (int i = t.getInterval().getLow(); i <= t.getInterval().getHigh(); i++) result[i] = t.getValue();
        }
        return result;
    }

}
